""" orthogonal.py

An example of orthogonal regions
"""
import sys
import unittest
import logging
import time
import enum

# setting path
sys.path.append('..')
from src.pyrarchical import hsm


class OrthogonalController:
    """ Example controller to manage continuous state for the orthogonal regions example """

    def __init__(self):
        self.action_history = []
        self.temperature = 0.0
        self.input = None

    def clear_action_history(self):
        self.action_history = []

    def record_action_history(self, action):
        self.action_history.append(action)

    def set_input(self, input):
        self.input = input


class TrackedState(hsm.State):
    """ States that record their actions in the controller for debugging and testing purposes """

    def init_action(self):
        super().init_action()
        self._controller.action_history.append(f'{self.name}_init')

    def entry_action(self):
        super().entry_action()
        self._controller.action_history.append(f'{self.name}_entry')

    def exit_action(self):
        super().exit_action()
        self._controller.action_history.append(f'{self.name}_exit')

    def during_action(self):
        super().during_action()
        self._controller.action_history.append(f'{self.name}_during')


class Events(enum.Enum):
    A = enum.auto()
    B = enum.auto()
    C = enum.auto()
    D = enum.auto()
    E = enum.auto()
    F = enum.auto()
    G = enum.auto()
    H = enum.auto()


class OrthogonalHsm(hsm.HSM):

    def __init__(self):
        # TODO: clean this up
        super().__init__()

        # First thing to do is to instantiate the controller for managing continuous state
        self._controller = OrthogonalController()

        # Add states by name
        self.add_region('left')  # HSM is implied parent
        self.add_region('right')  # orthogonal region names need to be unique across the HSM just like states
        self.add_state(TrackedState("s1"), parent='left', initial=True)
        self.add_state(TrackedState("s2"), parent='left')
        self.add_state(TrackedState("sA"), parent='right', initial=True)
        self.add_state(TrackedState("sB"), parent='right')
        self.add_region('fred', parent='s1')
        self.add_region('george', parent='s1')
        self.add_state(TrackedState("s11"), parent='fred', initial=True)
        self.add_state(TrackedState("s12"), parent='fred')
        self.add_state(TrackedState("s1X"), parent='george', initial=True)
        self.add_state(TrackedState("s1Y"), parent='george')

        # self.add_transition(source='s1', event=Events.E, target='sA') INVALID
        self.add_transition(source='sB', event=Events.A, target='sA')
        self.add_transition(source='sA', event=Events.B, target='sB')
        self.add_transition(source='s1X', event=Events.C, target='s2')
        self.add_transition(source='s1X', event=Events.D, target='s1')
        self.add_transition(source='s2', event=Events.E, target='s1')

        # TODO: clean this up
        self._initialize()


class OrthogonalTest(unittest.TestCase):

    def test_initial_conditions(self):
        orthogonal_hsm = OrthogonalHsm()

        # Active states provided as a nested dict by region
        initial_states = orthogonal_hsm.state
        self.assertEqual(initial_states['fred'].name, 's11')
        self.assertEqual(initial_states['george'].name, 's1X')
        self.assertEqual(initial_states['right'].name, 'sA')

        expected_actions = ['s1_entry', 's1_init', 's11_entry', 's1X_entry', 'sA_entry']
        self.assertEqual(orthogonal_hsm._controller.action_history, expected_actions)

    def test_transitions(self):
        """Test that a sequence of events go to the correct states."""
        orthogonal_hsm = OrthogonalHsm()

        # Event B in state sA
        orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        orthogonal_hsm.dispatch(Events.B)
        expected_actions = ['sA_exit', 'sB_entry']
        self.assertIs(orthogonal_hsm.state['right'].name, 'sB')
        self.assertEqual(orthogonal_hsm.state['fred'].name, 's11')  # unaffected
        self.assertEqual(orthogonal_hsm.state['george'].name, 's1X')  # unaffected
        self.assertEqual(orthogonal_hsm._controller.action_history, expected_actions)

        # Event A in state sB
        orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        orthogonal_hsm.dispatch(Events.A)
        expected_actions = ['sB_exit', 'sA_entry']
        self.assertIs(orthogonal_hsm.state['right'].name, 'sA')
        self.assertEqual(orthogonal_hsm.state['fred'].name, 's11')  # unaffected
        self.assertEqual(orthogonal_hsm.state['george'].name, 's1X')  # unaffected
        self.assertEqual(orthogonal_hsm._controller.action_history, expected_actions)

        # Event C in state s1X
        orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        orthogonal_hsm.dispatch(Events.C)
        expected_actions = ['s1X_exit', 's11_exit', 's1_exit', 's2_entry']
        self.assertEqual(orthogonal_hsm.state['left'].name, 's2')
        self.assertIs(orthogonal_hsm.state['right'].name, 'sA')  # unaffected
        self.assertEqual(orthogonal_hsm._controller.action_history, expected_actions)

        # Event E in state s2
        orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        orthogonal_hsm.dispatch(Events.E)
        expected_actions = ['s2_exit', 's1_entry', 's1_init', 's11_entry', 's1X_entry']
        self.assertEqual(orthogonal_hsm.state['fred'].name, 's11')
        self.assertEqual(orthogonal_hsm.state['george'].name, 's1X')
        self.assertIs(orthogonal_hsm.state['right'].name, 'sA')  # unaffected
        self.assertEqual(orthogonal_hsm._controller.action_history, expected_actions)

        # Event D in state s1X
        orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        orthogonal_hsm.dispatch(Events.D)
        # expected_actions = ['s1X_exit', 's11_exit', 's1_exit', 's1_entry', 's1_init', 's11_entry', 's1X_entry']  # external
        expected_actions = ['s1X_exit', 's11_exit', 's1_init', 's11_entry', 's1X_entry']  # local
        self.assertEqual(orthogonal_hsm.state['fred'].name, 's11')
        self.assertEqual(orthogonal_hsm.state['george'].name, 's1X')
        self.assertIs(orthogonal_hsm.state['right'].name, 'sA')  # unaffected
        self.assertEqual(orthogonal_hsm._controller.action_history, expected_actions)

    def test_speed(self):
        orthogonal_hsm = OrthogonalHsm()
        self.assertIs(orthogonal_hsm.state['right'].name, 'sA')
        self.assertEqual(orthogonal_hsm.state['fred'].name, 's11')
        self.assertEqual(orthogonal_hsm.state['george'].name, 's1X')

        num_cycles = 1E3
        num_transitions = 4
        start_time = time.time()
        for idx in range(int(num_cycles)):

            # Event B in state sA
            orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
            orthogonal_hsm.dispatch(Events.B)
            self.assertIs(orthogonal_hsm.state['right'].name, 'sB')

            # Event C in state s1X
            orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
            orthogonal_hsm.dispatch(Events.C)
            self.assertEqual(orthogonal_hsm.state['left'].name, 's2')

            # Event A in state sB
            orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
            orthogonal_hsm.dispatch(Events.A)
            self.assertIs(orthogonal_hsm.state['right'].name, 'sA')

            # Event E in state s2
            orthogonal_hsm._controller.clear_action_history()  # Reset the hsm continuous state
            orthogonal_hsm.dispatch(Events.E)
            self.assertEqual(orthogonal_hsm.state['fred'].name, 's11')
            self.assertEqual(orthogonal_hsm.state['george'].name, 's1X')

        stop_time = time.time()
        duration_ms = (stop_time - start_time) * 1E3
        print(f'Executed {int(num_cycles)} cycles of {num_transitions} in {duration_ms:.2f} ms')
        print(f'Average transition time = {duration_ms / num_cycles / num_transitions * 1E3:.2f} us')


##############################
#  Module Level Test Harness #
##############################
if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    # Kick off all unit tests contained in the file
    unittest.main()

    # test_case = OrthogonalTest()
    # test_case.test_transitions()



""" canonical.py

Re-implement Miro Samek's classic HSM example.

The purpose here is to explore different syntaxes for using this library.
"""
import sys
import unittest
import time
import enum

# setting path
sys.path.append('..')
from src.pyrarchical import hsm


class CanonicalController:
    """ Example controller to manage continous state for the canonical example """

    def __init__(self):
        self.action_history = []
        self.temperature = 0.0
        self.input = None

    def clear_action_history(self):
        self.action_history = []

    def record_action_history(self, action):
        self.action_history.append(action)

    def set_input(self, input):
        self.input = input


class CanonicalState(hsm.State):
    """ Define some basic behavior that is shared across all states

    Additional specific states could further derive from CanonicalState
    """

    def init_action(self):
        super().init_action()
        self._controller.action_history.append(f'{self.name}_init')

    def entry_action(self):
        super().entry_action()
        self._controller.action_history.append(f'{self.name}_entry')

    def exit_action(self):
        super().exit_action()
        self._controller.action_history.append(f'{self.name}_exit')

    def during_action(self):
        super().during_action()
        self._controller.action_history.append(f'{self.name}_during')


class Events(enum.Enum):
    A = enum.auto()
    B = enum.auto()
    C = enum.auto()
    D = enum.auto()
    E = enum.auto()
    E2 = enum.auto()
    F = enum.auto()
    G = enum.auto()
    H = enum.auto()


class CanonicalHsm(hsm.HSM):

    def __init__(self):
        # TODO: clean this up
        super().__init__()

        # First thing to do is to instantiate the controller for managing continuous state
        self._controller = CanonicalController()

        # Add states by name
        self.add_state(CanonicalState("s0"), initial=True)
        self.add_state(CanonicalState("s1"), parent='s0', initial=True)
        self.add_state(CanonicalState("s11"), parent='s1', initial=True)
        self.add_state(CanonicalState("s2"), parent='s0')
        self.add_state(CanonicalState("s21"), parent='s2', initial=True)
        self.add_state(CanonicalState("s211"), parent='s21', initial=True)
        self.add_state(CanonicalState("s212"), parent='s21')

        self.add_transition(source='s0', event=Events.E, target='s211')
        self.add_transition(source='s0', event=Events.E2, target='s212')

        self.add_transition(source='s1', event=Events.A, target='s1')
        self.add_transition(source='s1', event=Events.B, target='s11')
        self.add_transition(source='s1', event=Events.C, target='s2')
        self.add_transition(source='s1', event=Events.D, target='s0')
        self.add_transition(source='s1', event=Events.F, target='s211')

        self.add_transition(source='s11', event=Events.G, target='s211')

        self.add_transition(source='s2', event=Events.C, target='s1')
        self.add_transition(source='s2', event=Events.F, target='s11')

        self.add_transition(source='s21', event=Events.B, target='s211')
        self.add_transition(source='s21', event=Events.H, target='s21')

        self.add_transition(source='s211', event=Events.B, target='s21')
        self.add_transition(source='s211', event=Events.G, target='s0')
        self.add_transition(source='s211', event=None, target='s11', guard=self.s211_s11_guard)

        # TODO: clean this up
        self._initialize()

    def s211_s11_guard(self):
        try:
            return self._controller.input > 10.0
        except:
            return False


class HSMTest(unittest.TestCase):

    def test_events_states(self):
        """Test that a sequence of events go to the correct states."""

        # Check the start-up init and entry sequence to end in the right state
        # and having gone through the right sequence
        canonical_hsm = CanonicalHsm()
        u = ["s0_entry", "s0_init", "s1_entry", "s1_init", "s11_entry"]
        self.assertIs(canonical_hsm.state.name, 's11')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event a in state s11
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.A)
        # u = ["s11_exit", "s1_exit", "s1_entry", "s1_init", "s11_entry"]  # exteranl semantics
        u = ["s11_exit", "s1_init", "s11_entry"]  # internal semantics
        self.assertIs(canonical_hsm.state.name, 's11')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event e in state s11
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.E)
        u = ["s11_exit", "s1_exit", "s2_entry", "s21_entry", "s211_entry"]
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event e in state s211
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.E)
        u = ["s211_exit", "s21_exit", "s2_exit", "s2_entry", "s21_entry", "s211_entry"]  # internal path
        # u = []  # minimal path
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event a in state s211
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.A)
        u = []
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event h in state s211
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.H)
        # u = ["s211_exit", "s21_exit", "s21_entry", "s21_init", "s211_entry"]  # external semantics
        u = ["s211_exit", "s21_init", "s211_entry"]  # internal semantics
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event g in state s211
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.G)
        u = ["s211_exit", "s21_exit", "s2_exit", "s0_init", "s1_entry", "s1_init", "s11_entry"]  # local
        # external semantics
        # u = ["s211_exit", "s21_exit", "s2_exit", "s0_exit", "s0_entry", "s0_init", "s1_entry", "s1_init", "s11_entry"]
        self.assertIs(canonical_hsm.state.name, 's11')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event b in state s11
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.B)
        # u = ['s11_exit', 's1_exit', 's1_entry', 's11_entry']  # external
        u = ['s11_exit', 's11_entry']  # local
        # u = []  # minimal
        self.assertIs(canonical_hsm.state.name, 's11')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event g in state s11
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.G)
        u = ["s11_exit", "s1_exit", "s2_entry", "s21_entry", "s211_entry"]
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event b in state s211
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.B)
        u = ["s211_exit", "s21_init", "s211_entry"]
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event None in state s211, with input under 10.0
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(event=None, input=0.0)
        u = []
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event None in state s211, with input over 10.0
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(event=None, input=20.0)
        u = ["s211_exit", "s21_exit", "s2_exit", "s1_entry", "s11_entry"]
        self.assertIs(canonical_hsm.state.name, 's11')
        self.assertEqual(canonical_hsm._controller.action_history, u)

    def test_unknown_event(self):
        """An unknown event should fail somehow."""
        canonical_hsm = CanonicalHsm()
        with self.assertRaises(Exception, msg='Attempt use unknown event'):
            canonical_hsm.dispatch(Events.Z)

    def test_exit_only_to_least_common_ancestor(self):
        canonical_hsm = CanonicalHsm()

        # Event e in state s11
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.E)
        u = ["s11_exit", "s1_exit", "s2_entry", "s21_entry", "s211_entry"]
        self.assertIs(canonical_hsm.state.name, 's211')
        self.assertEqual(canonical_hsm._controller.action_history, u)

        # Event e2 in state s211
        canonical_hsm._controller.clear_action_history()  # Reset the hsm continuous state
        canonical_hsm.dispatch(Events.E2)
        u = ['s211_exit', 's21_exit', 's2_exit', 's2_entry', 's21_entry', 's212_entry']  # internal
        # u = ["s211_exit", "s212_entry"]  # minimal
        self.assertIs(canonical_hsm.state.name, 's212')
        self.assertEqual(canonical_hsm._controller.action_history, u)

    def test_events_entry_counter_simple(self):
        pass

    def test_events_entry_exit_init(self):
        pass

    def test_speed(self):
        hsm = CanonicalHsm()
        assert hsm.state.name == 's11'
        num_cycles = 1E3
        num_transitions = 2
        start_time = time.time()
        for idx in range(int(num_cycles)):
            hsm._controller.clear_action_history()  # Reset the hsm continuous state
            hsm.dispatch(Events.E)
            assert hsm.state.name == 's211'
            hsm.dispatch(Events.F)
            assert hsm.state.name == 's11'
        stop_time = time.time()
        duration_ms = (stop_time - start_time) * 1E3
        print(f'Executed {int(num_cycles)} cycles of {num_transitions} in {duration_ms:.2f} ms')
        print(f'Average transition time = {duration_ms / num_cycles / num_transitions * 1E3:.2f} us')


##############################
#  Module Level Test Harness #
##############################
if __name__ == "__main__":
    # Kick off all unit tests contained in the file
    unittest.main()



""" A simple example that uses the states-are-classes """

import enum
import sys
import logging

# setting path
sys.path.append('..')
from src.pyrarchical import hsm

logger = logging.getLogger(__name__)


class CdEvent(enum.Enum):
    PLAY = enum.auto()
    OPEN_CLOSE = enum.auto()
    STOP = enum.auto()
    CD_DETECTED = enum.auto()
    PAUSE = enum.auto()
    END_PAUSE = enum.auto()
    HAMMER = enum.auto()


class CdPlayer(hsm.HSM):
    """ The one-and-only top-level state (machine) """

    def __init__(self, controller):
        super().__init__(controller=controller)
        self._controller = controller

        # TODO: remove top state now that the HSM is the top state
        self.add_state(Top())
        self.set_initial('Top')

        # Final step
        self._initialize()


class Top(hsm.State):

    def __init__(self):
        super().__init__()

        self.add_state(NotBroken())
        self.add_state(Broken())
        self.set_initial('NotBroken')

        self.add_transition(event=CdEvent.HAMMER, target='Broken')


class Broken(hsm.State):
    """ Example showing the minimal amount of code required to define a state with this pattern """
    pass


class NotBroken(hsm.State):

    def __init__(self):
        super().__init__()
        self.add_state(Stopped())
        self.add_state(Open())
        self.add_state(Empty())
        self.add_state(Playing())
        self.add_state(Paused())
        self.set_initial('Empty')


class Stopped(hsm.State):

    def __init__(self):
        super().__init__()
        self.add_transition(event=CdEvent.PLAY, target='Playing')
        self.add_transition(event=CdEvent.OPEN_CLOSE, target='Open')
        self.add_transition(event=CdEvent.STOP, target=None, action=self._on_stop)

    def _on_stop(self):
        """ Just an example of how a state can handle an event without a transition """
        logger.info('Received redundant STOP event')

    def entry_action(self):
        """ Example showing how to create custom entry/exit actions """
        super().entry_action()
        logger.info('Displacy CD title screen')


class Open(hsm.State):

    def __init__(self):
        super().__init__()
        self.add_transition(event=CdEvent.OPEN_CLOSE, target='Empty')

    def entry_action(self):
        """ Example showing how to create custom entry/exit actions """
        super().entry_action()
        logger.info('Blank screen and display brand logo')


class Empty(hsm.State):

    def __init__(self):
        super().__init__()
        self.add_transition(event=CdEvent.OPEN_CLOSE, target='Open')
        self.add_transition(event=CdEvent.CD_DETECTED, target='Stopped')


class Playing(hsm.State):

    def __init__(self):
        super().__init__()
        self.add_transition(event=CdEvent.STOP, target='Stopped')
        self.add_transition(event=CdEvent.PAUSE, target='Paused')
        self.add_transition(event=CdEvent.OPEN_CLOSE, target='Open')


class Paused(hsm.State):

    def __init__(self):
        super().__init__()

        self.add_transition(event=CdEvent.STOP, target='Stopped')
        self.add_transition(event=CdEvent.PLAY, target='Playing')
        self.add_transition(event=CdEvent.OPEN_CLOSE, target='Open')


class DummyController:
    pass


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    controller = DummyController()
    cd_player = CdPlayer(controller)
    assert cd_player.state.name == 'Empty'

    cd_player.dispatch(CdEvent.OPEN_CLOSE)
    cd_player.execute_during_action()
    assert cd_player.state.name == 'Open'

    cd_player.dispatch(CdEvent.OPEN_CLOSE)
    cd_player.execute_during_action()
    assert cd_player.state.name == 'Empty'

    cd_player.dispatch(CdEvent.CD_DETECTED)
    cd_player.execute_during_action()
    assert cd_player.state.name == 'Stopped'

    cd_player.dispatch(CdEvent.PLAY)
    cd_player.execute_during_action()
    assert cd_player.state.name == 'Playing'

    cd_player.dispatch(CdEvent.PAUSE)
    cd_player.execute_during_action()
    assert cd_player.state.name == 'Paused'

    cd_player.dispatch(CdEvent.PLAY)
    cd_player.execute_during_action()
    assert cd_player.state.name == 'Playing'

    cd_player.dispatch(CdEvent.HAMMER)
    cd_player.execute_during_action()
    assert cd_player.state.name == 'Broken'



"""Heirarchical state machine module.

This module provides a framework for a responsive heirarchical state-machine.

For hybrid dynamical systems, the following update equations are commonly used.
```
x[k+1] = f_s(x[k], u(k]))
y[k] = g_s(x[k], u(k]))
```
where f and g are functions specific to each discrete state s, x is a
continuous state that exists in all discrete states, u is a set of inputs, and
y is the output of the system.

In a Python HSM implementation, only the during state will really obey the
discrete-time step, k. However, we may need to update states during other HSM
actions.
```
x = init_s(x, u)
x = entry_s(x, u)
x = during_s(x, u)
x = exit_s(x, u)
y = output_s(x, u)
```
The continuous state, x, is a member variable of the HSM. The output function
is a static method. The input, continuous state, and output each need an
object defined to describe them. At the simplest, this will just be a list.

The module will run a high level test if executed.
    $ python3 hsm.py

Copyright Matt Roelle 2018
"""

import logging
import enum
import copy

_log = logging.getLogger("hsm")


class Transition():
    """Transition in a heirarchical state-machine"""

    def __init__(self, name="null", event=None, target=None, action=None, guard=None):
        """Initialize the state"""
        self._name = name
        self.event = event
        self.target = target
        self.action = action
        self.guard = guard

    def __str__(self):
        return f'Transition name={self._name}, event={self.event}, target={self.target.name}'


class State():
    """State in a hierarchical state-machine"""

    def __init__(self, name=None):
        """Initialize the state"""
        if name is None:
            self._name = self.__class__.__name__  # TODO: make sure this is name of derived class!
        else:
            self._name = name
        self._controller = None  # controller is responsible for managing continous state
        self._state_machine = None  # the state machine that this state is part of
        self.parent = None
        self.initial = None
        self.transitions = []
        self._declared_transitions = []
        self._states = {}  # states can hold substates
        self._regions = {}  # for performance reasons, don't create a region until we have two!

    def __str__(self):
        return "State " + self._name

    @property
    def name(self):
        return self._name

    def set_controller(self, controller):
        """ Provide every state with a reference to a shared controller responsible for managing continuous state """
        self._controller = controller
        # recurse down into any states we might own as well
        for substate in self._states.values():
            substate.set_controller(controller)
        for region in self._regions.values():
            region.set_controller(controller)

    def set_state_machine(self, state_machine):
        """ Provide every state with a reference to the state machine which it belongs to """
        self._state_machine = state_machine
        # recurse down into any states we might own as well
        for substate in self._states.values():
            substate.set_state_machine(state_machine)
        for region in self._regions.values():
            region.set_state_machine(state_machine)

    def create_transition_object(self, event, target_object, action=None, guard=None):
        """ Intentionally obscure name to avoid confustion with user-facing add_transition function """
        self.transitions.append(Transition(event=event, target=target_object, action=action, guard=guard))

    def add_transition(self, event, target, action=None, guard=None):
        """ Declare a transition that will be instantiated after all states exist and are linked """
        self._declared_transitions.append((event, target, action, guard))

    def instantiate_declared_transitions(self):
        for declaration in self._declared_transitions:
            # TODO: used a named tuple for this
            self.create_transition_object(event=declaration[0],
                                          target_object=self.find_state_by_name(declaration[1]),
                                          action=declaration[2],
                                          guard=declaration[3])
        self._declared_transitions = []  # clear out the list to avoid the risk of duplicating these
        for name, substate in self._states.items():
            substate.instantiate_declared_transitions()
        for name, region in self._regions.items():
            region.instantiate_declared_transitions()

    def add_state(self, state, name=None, initial=False):
        """ Add a substate """
        # TODO: consider relaxing this requirement.  Can a superstate have multiple substates in the "default"
        #       region without calling it a region, and then also have one or more named regions?
        #       (probably not that hard to add once I figure out how to do this in the first place,
        #        but is additional complication I don't need right now)
        assert not self._regions, 'Cannot add substate directly to a superstate that already has regions'
        if name is None:
            name = state.name
        if name == 'null' or name in self._states:
            _log.error(f'state name {name} is either invalid or already exists')
            assert False
        # Store states in a name-indexed dictionary so that they can be retrieved from anywhere
        # Will likely want to link parents/children directly for performance reasons once all states are declared
        self._states[name] = state
        # When added directly to a state, the parent is obvious
        self._states[name].parent = self
        # Convenience option to simplify declaration syntax
        if initial:
            self.initial = state

    def add_region(self, name):
        """ Add an orthogonal region """
        assert not self._states, 'Regions must be declared before states'
        assert name not in self._regions, 'Region names must be unique'
        # Create the actual region
        self._regions[name] = Region(name)
        # Link parentage
        self._regions[name].parent = self

    def set_initial(self, initial):
        """ Define the initial (default) substate """
        self.initial = self._states[initial]

    def init_action(self):
        _log.info(f'Init {self._name}')

    def entry_action(self):
        _log.info(f'Enter {self._name}')

    def during_action(self):
        _log.info(f'During {self._name}')

    def exit_action(self):
        _log.info(f'Exit {self._name}')

    # CAUTION: There may not be a single top state.  You are probably looking for the state machine itself instead.
    # def get_top_state(self):
    #     """ Get a reference to the state at the very top """
    #     if self.parent is None:
    #         return self
    #     else:
    #         return self.parent.get_top_state()

    def find_substate_by_name(self, state_name):
        """ Find a sub-state (possible multiple levels deep) by name

        NOT to be used at run-time!
        """
        # search sub-states
        for name, state in self._states.items():
            if state.name == state_name:
                return state
            else:
                result = state.find_substate_by_name(state_name)
                if result is not None:
                    return result
        # search orthogonal regions (which can contain substates)
        for name, region in self._regions.items():
            if region.name == state_name:
                return region
            else:
                result = region.find_substate_by_name(state_name)
                if result is not None:
                    return result
        # We have finished searching all substates with no sucess
        return None

    def find_state_by_name(self, state_name):
        """ NOT to be use at run-time, only for initial transition linkage! """
        # most likely scenario is that we are looking for a substate
        if state_name is None:
            return None
        result = self.find_substate_by_name(state_name)
        if result is not None:
            return result
        # if that didn't work, jump to the top of the tree and try again
        result = self._state_machine.find_substate_by_name(state_name)
        if result is not None:
            return result
        # We can't find it!
        _log.error(f'Cannot find state {state_name}')
        assert False

    def first_child(self, state):
        """Recurse the parentage of a target state to find the first child."""
        # TODO: see if there is a more efficient way to do this now that
        #       we have direct access to all of our children
        if state is None:
            return None
        if state.parent is None:
            return None
        if state.parent is self:
            return state
        return self.first_child(state.parent)

    def find_target(self, event=None):
        """Recurse parental lineage until transition is found."""
        for transition in self.transitions:
            if event == transition.event:
                if transition.guard is None:
                    return self, transition.target
                elif transition.guard():
                    return self, transition.target
            if event is None and transition.guard is not None:
                if transition.guard():
                    return self, transition.target
        if isinstance(self.parent, State):
            return self.parent.find_target(event)
        return None, None

    def find_parent_region(self):
        """ Crawl upward looking for the next-containing region """
        if isinstance(self, Region):
            return self
        elif self.parent is None:
            return None
        else:
            return self.parent.find_parent_region()

    def seek_state(self, state, active_region_name):
        """ Work our way up or down the state hierarchy towards a goal state, entering and exiting as we go. """
        if self is state:
            # We have gotten to where we are going
            return active_region_name
        child = self.first_child(state)
        if child is not None:
            # we found a path to the goal state through one of our children, so head that way

            if self._regions:
                # We are getting ready to seek down into an orthogonal region, so we need to also
                # enter/init all the other regions
                # We only keep "leaf" regions in HSM.__state_vector, so whenever we descend into new regions,
                # we have to remove (deactivate) the old (parent) region
                self._state_machine.deactivate_region(region_name=active_region_name)
                for region_name, region in self._regions.items():
                    self._state_machine.activate_region(region_name=region_name, current_state=region)
                    if region is not child:
                        region.entry_action()
                        region.run_init(active_region_name=region_name)
                active_region_name = child.name

            child.entry_action()
            self._state_machine.set_current_state(region_name=active_region_name, current_state=child)
            return child.seek_state(state=state, active_region_name=active_region_name)

        elif self.parent is not None:
            # go up the tree asking the same question
            _log.debug(f"{self} go up to " + str(self.parent))

            # # This might be redundant with having exited the region below
            # if self._regions:
            #     # We are exiting a state that has orthogonal regions, so we need to exit ALL of them
            #     self._exit_all_regions()  # this will deactivate all regions as they exit
            #     active_region_name = self.find_parent_region().name
            #     if not self._state_machine.is_region_active(active_region_name):
            #         self._state_machine.activate_region(region_name=active_region_name, current_state=self)

            self.exit_action()

            if isinstance(self, Region):
                self._state_machine.deactivate_region(region_name=active_region_name)

                # We are exiting a region (we may or may not be exiting the parent state next) and
                # need to exit the sibling states.. This will deactivate all regions as they exit
                # and activate the new parent region upon completion
                active_region_name = self.parent._exit_all_regions()  # this will deactivate all regions as they exit

            self._state_machine.set_current_state(region_name=active_region_name, current_state=self.parent)
            return self.parent.seek_state(state=state, active_region_name=active_region_name)
        else:
            # we reached the top of the tree.  oops!
            _log.error(f'Seek to {state} failed at {self}')
            # TODO: find a more gracious way to fail
            assert False

    def _exit_all_regions(self):
        """ Ensure that we have exited (and deactivated) all regions of this state (which may not yet be the current state) """
        _log.debug(f'{self}._exit_all_regions()')
        for region_name, region in self._regions.items():
            # tell each leaf state in any of our regions to exit up to the current level (catch up)
            if self._state_machine.is_region_active(region_name):
                # The immediate sibling region is still active (it has not been further split, or already deactivated),
                # so we can easily look up its current leaf state and tell it to exit
                self._state_machine.state[region_name].seek_state(self, active_region_name=region_name)
                # the region should deactivate itself when it exits up to the current state level
            else:
                # TODO: we need to go track down the sub-regions
                continue

        new_active_region = self.find_parent_region().name
        self._state_machine.activate_region(region_name=new_active_region, current_state=self)
        return new_active_region

    # TODO: move this to Hsm class?
    def seek_lowest_common_ancestor(self, target_state):
        current_state = self
        while current_state is not target_state and current_state.parent is not None:
            # do I have an immediate child that is an ancestor of target_state?
            child = current_state.first_child(target_state)
            if child is None:  # first_child search reached top of state machine
                _log.debug("Go up to " + str(current_state.parent))
                current_state.exit_action()
                current_state = current_state.parent
            else:  # found a child of current_state that leads to target state
                return current_state
        return current_state

    def run_init(self, active_region_name):
        _log.debug("Run the init for " + str(self))
        if self._states:  # we have substates, so descend into the initial state
            self.init_action()
            self.initial.entry_action()
            self._state_machine.set_current_state(region_name=active_region_name, current_state=self.initial)
            # recurse down into the substate
            self.initial.run_init(active_region_name=active_region_name)
        elif self._regions:  # TODO: relax assumptions that regions and substates are mutually exclusive?
            self.init_action()
            # We only keep "leaf" regions in HSM.__state_vector, so whenever we descend into new regions,
            # we have to remove (deactivate) the old (parent) region
            self._state_machine.deactivate_region(region_name=active_region_name)
            for region_name, region in self._regions.items():
                region.entry_action()
                self._state_machine.activate_region(region_name=region_name, current_state=region)
                region.run_init(active_region_name=region_name)
        else:  # we found the leaf
            pass


class Region(State):
    """ An orthogonal region

    Modeling this after a state itself may for now, hoping for the best....

    If this were C++, this would be a PRIVATE inheritance.
    I want to re-use the implementation, but not to be considered a state externally (by interface)
    """
    def __init__(self, name=None):
        """Initialize the orthogonal region just the same as we would a state"""
        super().__init__(name=name)
        self._active_subregion_names = []  # in case this region gets further split at some point

    def __str__(self):
        return "Region " + self._name


class HSM(State):

    def __init__(self, controller=None):
        super().__init__()
        self.__state_vector = {}  # What is the "active" state?  One entry per active orthogonal region!
        self._controller = controller

    def _initialize(self):
        """Initialize the hierarchical state machine."""
        # Ensure that all states have an active link to the controller
        if self._controller is not None:
            for state in self._states.values():
                state.set_controller(self._controller)
            for region in self._regions.values():
                region.set_controller(self._controller)

        # Ensure that all states have a link back to the state machine
        self.set_state_machine(self)  # inherited from State, will recurse downward

        # Substates might have transitions declared that still need to be instantiated
        # now that all states have been instantiated and linked
        # We do this as a two-step process to allow flexibilty in transtion declaration
        # and maintain the runtime performance of transition objects containing references
        # to the actual target state object.
        for state in self._states.values():
            state.instantiate_declared_transitions()
        for region in self._regions.values():
            region.instantiate_declared_transitions()

        if self.initial:
            self.activate_region(region_name='default', current_state=self.initial)
        elif self._regions:
            for name, region in self._regions.items():
                self.activate_region(region_name=name, current_state=region)
        else:
            _log.error('Expected to find either an initial state or orthogonal regions')

        # We have to iterate through a copy of the state vector since it may change size during init()
        # I am reasonably certain that all the initial members  of the dict remain valid at least
        # until they are called in the loop below (there is no cross-talk between branches of the tree)
        state_vector_copy = copy.copy(self.__state_vector)  # SHALLOW copy!!!
        for region_name, state in state_vector_copy.items():  # for implementation purposes, a region IS A state
            state.entry_action()
            # states are now responsible for updating HSM.__state through function calls to HSM
            state.run_init(active_region_name=region_name)

    def add_state(self, state, name=None, parent=None, initial=False):
        """ Try to encourage users toward adding states through this function to keep implementation private """
        # Add substates directly to their parent states
        # Yes, this does for a declaration order, but I think it will help with orthogonal regions
        # if states always "belong" to their superstate (or region)
        if parent is not None:
            parent_state = self.find_substate_by_name(parent)
            parent_state.add_state(state=state, name=name, initial=initial)
        else:
            # add state directly to HSM (top-level)
            super().add_state(state=state, name=name, initial=initial)

    def add_region(self, name, parent=None):
        """ Try to encourage users toward adding states through this function to keep implementation private """
        # Add substates directly to their parent states
        # Yes, this does for a declaration order, but I think it will help with orthogonal regions
        # if states always "belong" to their superstate (or region)
        if parent is not None:
            parent_state = self.find_substate_by_name(parent)
            parent_state.add_region(name=name)
        else:
            # add region directly to HSM (top-level)
            super().add_region(name=name)

    def set_initial(self, initial, source=None):
        """ Define the initial (default) substate """
        if source is not None:
            source_state = self.find_state_by_name(source)
            source_state.set_initial(initial)
        else:
            super().set_initial(initial)

    def activate_region(self, region_name, current_state):
        """ Called by states as they descend into or exit from orthogonal regions """
        _log.info(f'Activating {region_name} region with current state {current_state.name}')
        self.__state_vector[region_name] = current_state

    def deactivate_region(self, region_name):
        """ Called by states as they descend into or exit from orthogonal regions """
        _log.info(f'Deactivating {region_name} region')
        del self.__state_vector[region_name]

    def is_region_active(self, region_name):
        """ Is the specified region currently active (it has a leaf state) """
        return region_name in self.__state_vector

    # QUESTION: Do we prefer "current state" or "active state"?
    def set_current_state(self, region_name, current_state):
        """ Called by states as they crawl up or down the hierarchy """
        self.__state_vector[region_name] = current_state

    def add_transition(self, event, target, action=None, guard=None, source=None):
        """ Add a transition to the state machine

        Overloads State.add_transition() by including a soucre
        """
        # INTERNALLY, transitions are stored as part of the source state
        if source is not None:
            source_state = self.find_substate_by_name(source)
            source_state.add_transition(event=event, target=target, action=action, guard=guard)
        else:
            super().add_transition(event=event, target=target, action=action, guard=guard)

    def dispatch(self, event, input=None):
        """Process events with the state machine."""

        # for now, just process the orthogonal regions in whatever order they happen to be in
        # TODO: dedicated function to loop through leaf states
        state_vector_copy = copy.copy(self.__state_vector)  # SHALLOW copy!!!
        for region_name, state in state_vector_copy.items():
            if event is not None:
                _log.info(f'dispatching {event} in {state} in {region_name} region')
            else:
                _log.debug(f'checking {state} in {region_name} region for auto transitions')
            if input is not None and self._controller is not None:
                self._controller.set_input(input)
            event_state, target_state = state.find_target(event)
            if target_state is None:  # No transitions found  TODO: handle local transitions (which have no target)
                _log.debug('No transition found')
                continue  # don't break out of the for loop, just go check the next region

            _log.debug("Go to the state with the transition : " + str(event_state))
            # The UML spec is surprisingly difficult to interpret here, but I am now (mostly)
            # convinced that you are supposed to exit all the way up to the level at which
            # the transition is DEFINED, even if there is a shorter path to the target from
            # your current derived state.
            active_region_name = state.seek_state(state=event_state, active_region_name=region_name)
            # This implements the "minimal" pathway, which may come back as a non-standard option later
            # See https://gitlab.com/nealtanner/pyrarchical-state-machine/-/issues/5
            # self.__state_vector[region_name] = self.__state_vector[region_name].seek_lowest_common_ancestor(target_state)

            _log.debug("Transition from here : " + str(self.__state_vector[region_name]))
            # N.B.  Seeking up to the event state above could have changed what region we are in
            active_region_name = self.__state_vector[active_region_name].seek_state(
                state=target_state, active_region_name=active_region_name)

            _log.debug(f"Run init for {self.__state_vector[active_region_name]}")
            self.__state_vector[active_region_name].run_init(active_region_name=active_region_name)

            _log.debug("Check for auto-transitions")
            self.dispatch(event=None)

    @property
    def state(self):
        """Get the active state."""
        # to keep most users blissfully unaware of orthogonal regions...
        if len(self.__state_vector) == 1:
            return self.__state_vector['default']
        else:
            return self.__state_vector

    def execute_during_action(self):
        for region_name, state in self.__state_vector.items():
            state.during_action()


if __name__ == "__main__":

    class Event(enum.Enum):
        RESET = enum.auto()
        FLIP = enum.auto()
        FLOP = enum.auto()

    class Example(HSM):

        def __init__(self):
            # TODO: clean this up
            super().__init__()

            # Add states by name
            self.add_state(State("flipped"), initial=True)
            self.add_state(State("flopped"))

            self.add_transition(event=Event.RESET, target='flipped')
            self.add_transition(source='flipped', event=Event.FLOP, target='flopped')
            self.add_transition(source='flopped', event=Event.FLIP, target='flipped')

            self._initialize()

    # logging.basicConfig(level=logging.INFO)
    logging.basicConfig(level=logging.INFO)
    # logging.basicConfig(filename="hsm.log", filemode="w", level=logging.INFO)

    example_hsm = Example()

    example_hsm.dispatch(Event.RESET)
    assert example_hsm.state.name == 'flipped'
    example_hsm.dispatch(Event.FLOP)
    assert example_hsm.state.name == 'flopped'
    example_hsm.dispatch(Event.FLIP)
    assert example_hsm.state.name == 'flipped'



""" diagram.py

Injest a pyrarchical-state-machine and draw diagrams of it with plant UML.

This code shamelessly borrowed from https://github.com/MainspringEnergy/eta-hsm

I originally thought that the functionality to spit out plant UML code would just be build directly
into hsm.py, but to keep that code lightweight, I am goin to start off with using this as an external
translation module.
"""
import hsm

class Transition:
    """An event that causes a transition to another state.

    possibly subject to a guard.
    """

    def __init__(self, source, events, target, guard=None, action=None):
        """Give up and just making these public for now."""
        self.source = source
        self.events = events  # could be a string or a list/tuple
        self.target = target
        self.guard = guard
        self.action = action

    def events_as_string(self, wrap=None):
        """Get the one or several events as a printable string."""
        if isinstance(self.events, str):
            return self.events
        elif len(self.events) == 1:
            return self.events[0]
        else:
            event_string = "("
            for idx, event in enumerate(self.events):
                event_string += event
                if idx < (len(self.events) - 1):  # there are more to come
                    if wrap and len(self.events) > wrap:  # there are too many for a single line
                        event_string += "\\n| "
                    else:
                        event_string += " | "
            event_string += ")"
            return event_string

    def euml_string(self):
        """Describe the transition in eUML syntax."""
        return "{} + {} [{}] / {} == {}".format(
            self.source,
            self.events_as_string(),
            self.guard,
            self.action,
            self.target,
        )


class State:
    """Container to represent a state in the state machine."""

    def __init__(self, name, parent=None):
        """Initialize a new state."""
        self._name = name
        self._parent = parent
        self._children = []
        self._initial_state = None
        self._transitions = []
        self._entry = []
        self._exit = []
        # children are responsible for registering themselves
        # with their parents
        if self._parent:
            self._parent.register_child(self)

    @property
    def name(self):
        """Read-only access to name."""
        return self._name

    @property
    def parent(self):
        """Read-only access to parent."""
        return self._parent

    @property
    def children(self):
        """Read-only access to children."""
        return self._children

    @property
    def entry(self):
        """Entry property."""
        return self._entry

    @property
    def exit(self):
        """Exit property."""
        return self._exit

    @property
    def initial_state(self):
        """Initialize property."""
        return self._initial_state

    @property
    def transitions(self):
        """Transitions property."""
        return self._transitions

    def is_substate_of(self, name):
        """Is substrate of.

        Is this state a substate of some other named state?
        """
        if self.parent.name == name:
            return True
        elif self.parent.name is None:
            return False
        else:
            # recurse our way up the state hierarchy
            return self.parent.is_substate_of(name)

    def levels_of_nesting_below(self, name):
        """Level of nesting below name.

        How many levels deep is this state nested below the named super-state?
        """
        # print('{}.levels_of_nesting_below({})'.format(self.name, name))
        if self.name == "Top":
            return 0
        elif self.name == name:
            return 0
        elif self.parent.name == name:
            return 1
        elif self.parent.name == "Top":
            return None  # to signify failure
        else:
            # recurse our way up the state hierarchy
            return self.parent.levels_of_nesting_below(name) + 1

    def ancestor_at_most_n_levels_of_nesting_below(self, name, max_depth):
        """Ancestor at most n levels of nesting below.

        Find ancestor that is at most N levels of nesting below named
        super-state.
        """
        assert self.is_substate_of(name)
        if self.levels_of_nesting_below(name) <= max_depth:
            return self
        else:
            return self.parent.ancestor_at_most_n_levels_of_nesting_below(
                name, max_depth
            )

    def register_child(self, child):
        """Register child.

        Children are responsible for registering themselves with the parent
        """
        assert child.parent == self
        self._children.append(child)

    def add_initial_state(self, initial_state):
        """Add initial state.

        Initial state transition for a composite state is represented by
        a dot on the diagram.
        """
        assert initial_state in self._children
        self._initial_state = initial_state

    def add_transition(self, transition):
        """Add transition.

        Transitions are defined by an event, target_state, and guard_condition.
        """
        assert transition.source == self._name
        self._transitions.append(transition)

    def add_entry(self, statement):
        """Add entry.

        For now, we're just going to keep the full block of code
        we found in hsmEntry().
        """
        stripped_statement = statement.strip()
        if stripped_statement[:2] == "//":
            # ignore commented out lines
            return
        if stripped_statement == "":
            # ignore empty lines
            return
        self._entry.append(stripped_statement)

    def add_exit(self, statement):
        """Add exit.

        For now, we're just going to keep the full block of code
        we found in hsmExit().
        """
        stripped_statement = statement.strip()
        if stripped_statement[:2] == "//":
            # ignore commented out lines
            return
        if stripped_statement == "":
            # ignore empty lines
            return
        self._exit.append(stripped_statement)

    def print_state_hierarchy(self, indentation=0):
        """Print state hierarchy."""
        description_of_self = " " * indentation + self._name
        if isinstance(self, Region):
            description_of_self += " (region)"
        if self._initial_state:
            description_of_self += " ( --> {})".format(self._initial_state.name)
        print(description_of_self)
        for child in self._children:
            child.print_state_hierarchy(indentation=indentation + 4)

    def print_transition_table(self, indentation=0):
        """Print transition table."""
        for transition in self._transitions:
            print(" " * indentation + transition.euml_string())
        for child in self._children:
            child.print_transition_table(indentation=indentation + 4)


class Region(State):
    pass


class StateMachine:
    """A collection of states makes a state machine."""

    def __init__(self, hsm):
        """Initialize a new state machine."""
        self._hsm = hsm
        self._basename = hsm.name
        self._states = {}
        self._plant_uml_legend = []
        self._plant_uml_out_of_scope_transition = False

    def state(self, name):
        """Get a reference to an individual state by name."""
        return self._states[name]

    def add_state(self, state):
        """Add state."""
        self._states[state.name] = state
        # TODO: is there any additional linking we need to do here?
        # TODO: Maybe populate children?

    def _add_legend_entry(self, entry):
        """Add legend entry.

        Legend entries to appear at the bottom of the plantUML diagram.
        """
        if entry in self._plant_uml_legend:
            # already have this one, do not repeat
            return
        else:
            self._plant_uml_legend.append(entry)

    def print_state_hierarchy(self, top=None):
        """Print state hierarchy."""
        if top is None:
            if self._basename in self._states:
                top = self._basename
            else:
                top = 'Top'  # Hope that we have a top state called 'Top'
        print("State hierarchy for {} state machine".format(self._basename))
        self.state(top).print_state_hierarchy()

    def print_transition_table(self, top=None):
        """Print transition table."""
        if top is None:
            if self._basename in self._states:
                top = self._basename
            else:
                top = 'Top'  # Hope that we have a top state called 'Top'
        print("Transition table for {} state machine".format(self._basename))
        self.state(top).print_transition_table()

    def extract_everything(self):
        """Extract everything.

        Scan files in standard locations for all of the things
        we know how to scrape.
        """
        self.extract_state_tree()
        self.extract_initial_states()
        self.extract_transitions()
        self.extract_entry_exit_actions()

    def extract_state_tree(self, hsm_state=None):
        """ Extract state tree from a pyrarchical-state-machine """
        if hsm_state is None:
            hsm_state = self._hsm
        print(f'extract_state_tree({hsm_state.name})')

        if isinstance(hsm_state, hsm.Region):
            self.add_state(Region(name=hsm_state.name, parent=self.state(hsm_state.parent.name)))
        else:
            if hsm_state.parent:
                self.add_state(State(name=hsm_state.name, parent=self.state(hsm_state.parent.name)))
            else:
                self.add_state(State(name=hsm_state.name))

        # recurse down into substates
        for substate_name, substate in hsm_state._states.items():
            self.extract_state_tree(hsm_state=substate)

        for region_name, region in hsm_state._regions.items():
            self.extract_state_tree(hsm_state=region)

    def extract_initial_states(self, hsm_state=None):
        """ Extract initial states.

        Performed as a separate recursion to ensure states exist already.
        """
        if hsm_state is None:
            hsm_state = self._hsm
        print(f'extract_initial_states({hsm_state.name})')

        if hsm_state.initial:
            self.state(hsm_state.name).add_initial_state(self.state(hsm_state.initial.name))

        # recurse down into substates
        for substate_name, substate in hsm_state._states.items():
            self.extract_initial_states(hsm_state=substate)

        for region_name, region in hsm_state._regions.items():
            self.extract_initial_states(hsm_state=region)

    def extract_transitions(self, hsm_state=None):
        """ Extract transitions """
        if hsm_state is None:
            hsm_state = self._hsm
        print(f'extract_transitions({hsm_state.name})')

        for hsm_transition in hsm_state.transitions:
            if hsm_transition.event:
                event = hsm_transition.event.name
            else:
                event = None
            transition = Transition(source=hsm_state.name,
                                    events=[event],
                                    guard='omitted guard' if hsm_transition.guard else None,
                                    target=hsm_transition.target.name if hsm_transition.target else None)
            # print(transition.euml_string())
            self.state(hsm_state.name).add_transition(transition)

        # recurse down into substates
        for substate_name, substate in hsm_state._states.items():
            self.extract_transitions(hsm_state=substate)

        for region_name, region in hsm_state._regions.items():
            self.extract_transitions(hsm_state=region)

    def extract_entry_exit_actions(self, hsm_state=None):
        """ Extract entry exit actions. """

        raise NotImplementedError

        # self.state(state_name).add_entry(line)
        # self.state(state_name).add_exit(line)

    def generate_event_set(self):
        """Generate a set of all events that are USED by the state machine.

        This conceptually could be different than the list of events that are
        DEFINED in the enum.
        """
        events = set()  # unordered and unique
        for name, state in self._states.items():
            for transition in state.transitions:
                events.update(transition.events)
        return events

    def generate_plant_uml(self, top=None, max_depth=None, do_not_expand=None,
                           include_actions=True, include_guards=True,
                           filename=None):
        """Generate plant uml.

        Generate the text input expected by PlantUML for automatically drawing a UML state diagram.

        `top` specifies the top of **this** diagram, effectively defining the scope
        `do_not_expand` is a LIST of states to not expand into substates
        """
        if top is None:
            top = self._basename

        # Clear members that will be populated by the recursive calls
        # to _generate_plant_uml_for_state below
        self._plant_uml_legend = []
        self._plant_uml_out_of_scope_transition = False

        with open(filename, "w") as fid:
            # header
            fid.write("@startuml\n")

            # First pass:  Just declare states
            self._generate_plant_uml_states_for_state(fid, state=self.state(top), top=top,
                                               max_depth=max_depth, do_not_expand=do_not_expand)

            # If any transition **targets** are out of the current diagram scope, we create a single explicitly named
            # fake-state at the top level for these transitions to point to.
            if self._plant_uml_out_of_scope_transition:
                fid.write("state OutOfScope {\n}\n")

            # Second pass: Declare transitions
            self._generate_plant_uml_transitions_for_state(fid, state=self.state(top), top=top,
                                               max_depth=max_depth, do_not_expand=do_not_expand,
                                               include_actions=include_actions, include_guards=include_guards)

            if self._plant_uml_legend:
                fid.write("legend\n")
                for entry in self._plant_uml_legend:
                    fid.write("  {}\n".format(entry))
                fid.write("end legend\n")

            # footer
            fid.write("@enduml\n")

    def _generate_plant_uml_states_for_state(self, fid, state, top, max_depth, do_not_expand,
                                             include_actions=False, include_guards=False,
                                             indentation=0):
        """ declare sub-states (but do not decend into them yet)

        plant-uml seems to be creating local substates for any transition terminations that are not yet defined,
        which seems wrong.  this is an attempt to work around that by declaring states earlier in the sequence.
        it is likely a band-aid at best.
        """
        initial_indent = " " * indentation
        internal_indent = " " * (indentation + 2)

        # declare ourselves as a state
        if not isinstance(state, Region):
            fid.write(initial_indent + f'state {state.name} {{\n')

        # describe transition to initial state, if any
        if do_not_expand is not None and state.name in do_not_expand:
            # regardless of how deep we are, we will not descend into our sub-states
            pass
        elif max_depth is None or state.levels_of_nesting_below(top) < max_depth:
            if state.initial_state:
                fid.write(internal_indent + "[*] --> {}\n".format(state.initial_state.name))

        # entry/exit actions
        if include_actions:  # even if at max_depth
            for entry in state.entry:
                fid.write(internal_indent + "{} : entry / {}\n".format(state.name, entry))
            for exit in state.exit:
                fid.write(internal_indent + "{} : exit / {}\n".format(state.name, exit))

        # descend into sub-states
        if do_not_expand is not None and state.name in do_not_expand:
            # regardless of how deep we are, we will not descend into our sub-states
            pass
        elif max_depth is None or state.levels_of_nesting_below(top) < max_depth:
            for child in state.children:
                if isinstance(child, Region):
                    if 'already_defined_first_region' in locals():
                        # insert separator between regions
                        fid.write(internal_indent + "--\n")
                    else:
                        already_defined_first_region = True

                self._generate_plant_uml_states_for_state(fid, state=child, top=top,
                                                          max_depth=max_depth, do_not_expand=do_not_expand,
                                                          include_actions=include_actions, include_guards=include_guards,
                                                          indentation=indentation + 2)
        # close our block
        if isinstance(state, Region):
            # Have to declare transitions before we leave the region
            self._generate_plant_uml_transitions_for_state(fid=fid, state=state, top=top,
                                                           max_depth=max_depth, do_not_expand=do_not_expand,
                                                           include_actions=include_actions, include_guards=include_guards,
                                                           indentation=indentation + 1)
        else:
            fid.write(initial_indent + "}\n")

    def _generate_plant_uml_transitions_for_state(self, fid, state, top, max_depth,
                                      do_not_expand, include_actions, include_guards,
                                      indentation=0):
        """Generate plant uml for state.

        Generate the text input expected by PlantUML for automatically drawing a UML state diagram

        `top` specifies the top of **this** diagram, effectively defining the scope.

        Transitions to states outside of this scope will be directed to UML's "final" state.

        Transitions from states outside of this scope are (for now) simply omitted.

        This functionality original resided within the State class, but I was having to pass in a reference
        to the state machine in order to look up "scope" of target states, so I decided to move it here to the
        StateMachine class.  The downside is that now I had to make more of the details of the State class public.
        """
        initial_indent = " " * indentation

        # list transitions
        for transition in state.transitions:
            arrow = "-->"

            # figure out if the target state is hidden by do_not_expand list
            do_not_expand_state_hiding_target = None
            if do_not_expand is not None:
                for name in do_not_expand:
                    if self.state(transition.target).is_substate_of(name):
                        do_not_expand_state_hiding_target = name

            # catch transitions to states that are out of scope for the
            # current diagram
            if transition.target is None:
                print(f'ignoring self-transition in {transition.source}')
            elif not self.state(transition.target).is_substate_of(top):
                # Previously useed UML's "Final" state icon to represent that we have left the current diagram.
                # The problem with this was that it generated a "Final" icon locally in every state that
                # had an out-of-scope transition, which often bloated the diagram.  Instead, we're going to
                # create a single explicit "OutOfScope" state for all of these transitions to point to.
                diagram_target = "OutOfScope"  # '[*]'
                # Flag that this occurred so that generate_plant_uml can create the state at the top level
                self._plant_uml_out_of_scope_transition = True
            elif (max_depth is not None
                  and state.levels_of_nesting_below(top) == max_depth
                  and self.state(transition.target).is_substate_of(state.name)):
                # If the current state is already the max depth, then drop
                # transitions to our own substates (instead of letting the
                # next elif redirect them to ourselves)
                break
            elif do_not_expand is not None and state.name in do_not_expand:
                # If the current state is explicitly not going to be expanded,
                # then drop transitions to our own substates so that they do
                # not get auto-drawn.
                break
            elif max_depth is not None and self.state(transition.target).levels_of_nesting_below(top) > max_depth:
                # redirect transitions to deeply nested states to their parent state
                diagram_target = self.state(transition.target).ancestor_at_most_n_levels_of_nesting_below(top, max_depth).name
                # use a different style arrow to explicitly show that this has
                # been modified/approximated
                arrow = "-[dotted]->"
                self._add_legend_entry("dotted arrow = transition to hidden substate")
            elif do_not_expand_state_hiding_target:
                # redirect transitions to states hidden by do_no_expand list
                diagram_target = do_not_expand_state_hiding_target
                # use a different style arrow to explicitly show that this has
                # been modified/approximated
                arrow = "-[dotted]->"
                self._add_legend_entry("dotted arrow = transition to hidden substate")
            else:  # target in scope
                diagram_target = transition.target

            # start generating the actual plantUML string to represent a transition
            transition_string = f'{transition.source} {arrow} {diagram_target} : {transition.events_as_string(wrap=3)}'
            if transition.guard:
                if include_guards:
                    transition_string += " [{}]".format(transition.guard)
                else:
                    # explicitly show that a guard condition has been omitted
                    transition_string += " [*]"
                    self._add_legend_entry("[*] = guard condition omitted")

            if transition.action:
                if include_actions:
                    transition_string += " / {}".format(transition.action)
                else:
                    # explicitly show that a transition action has been omitted
                    transition_string += " /*"
                    self._add_legend_entry("/* = transition action omitted")

            fid.write(initial_indent + transition_string + "\n")

        # descend into sub-states
        if do_not_expand is not None and state.name in do_not_expand:
            # regardless of how deep we are, we will not descend into our sub-states
            pass
        elif max_depth is None or state.levels_of_nesting_below(top) < max_depth:
            for child in state.children:
                if not isinstance(child, Region):
                    self._generate_plant_uml_transitions_for_state(fid, state=child, top=top,
                                                       max_depth=max_depth, do_not_expand=do_not_expand,
                                                       include_actions=include_actions, include_guards=include_guards,
                                                       indentation=indentation + 2)


if __name__ == "__main__":
    import examples.cd_player
    import examples.canonical
    import examples.orthogonal

    # -------- CdPlayer -----------
    controller = examples.cd_player.DummyController()
    cd_player_hsm = examples.cd_player.CdPlayer(controller)
    state_machine = StateMachine(hsm=cd_player_hsm)
    state_machine.extract_state_tree()
    state_machine.extract_initial_states()
    state_machine.extract_transitions()

    print()
    state_machine.print_state_hierarchy()
    print()
    state_machine.print_transition_table()

    # default plantUML output
    state_machine.generate_plant_uml(filename="../../diagrams/cd_player.txt")

    # -------- Canonical -----------
    canonical_hsm = examples.canonical.CanonicalHsm()
    state_machine = StateMachine(hsm=canonical_hsm)
    state_machine.extract_state_tree()
    state_machine.extract_initial_states()
    state_machine.extract_transitions()

    print()
    state_machine.print_state_hierarchy()
    print()
    state_machine.print_transition_table()

    # default plantUML output
    state_machine.generate_plant_uml(filename="../../diagrams/canonical.txt")

    # -------- Orthogonal -----------
    orthogonal_hsm = examples.orthogonal.OrthogonalHsm()
    state_machine = StateMachine(hsm=orthogonal_hsm)
    state_machine.extract_state_tree()
    state_machine.extract_initial_states()
    state_machine.extract_transitions()

    print()
    state_machine.print_state_hierarchy()
    print()
    state_machine.print_transition_table()

    # default plantUML output
    state_machine.generate_plant_uml(filename="../../diagrams/orthogonal.txt")
